#pragma once

static const int VRT_VERSION_1  =  1;	// Original version
static const int VRT_VERSION_2  =  2;	// Added ability to show IC pin numbers
static const int VRT_VERSION_3  =  3;	// Added ground fill parameters
static const int VRT_VERSION_4  =  4;	// Simplified mapping of pinChar to pinIndex
static const int VRT_VERSION_5  =  5;	// Added flag to show grid
static const int VRT_VERSION_6  =  6;	// Added color saturation
static const int VRT_VERSION_7  =  7;	// Added pin labels
static const int VRT_VERSION_8  =  8;	// Added crystal component type
static const int VRT_VERSION_9  =  9;	// Added more trimmers/relays/switches
static const int VRT_VERSION_10 = 10;	// Added rectangles
static const int VRT_VERSION_11 = 11;	// Added component manager, bounding rects, reworked Footprint class
static const int VRT_VERSION_12 = 12;	// Added horizontal vero strips
static const int VRT_VERSION_13 = 13;	// Added crop margin
static const int VRT_VERSION_14 = 14;	// Added text manager
static const int VRT_VERSION_15 = 15;	// Added text color
static const int VRT_VERSION_16 = 16;	// Added more components
static const int VRT_VERSION_17 = 17;	// Added text sizes for components and pins
static const int VRT_VERSION_18 = 18;	// Added custom component type and shapes class
static const int VRT_VERSION_19 = 19;	// Added component editor flag and new component attributes
static const int VRT_VERSION_20 = 20;	// Added SURFACE_HOLE attribute
static const int VRT_VERSION_21 = 21;	// Added target board size
static const int VRT_VERSION_22 = 22;	// Label offsets measured in 1/16th of a square instead of 1/4
static const int VRT_VERSION_23 = 23;	// Added component definer
static const int VRT_VERSION_24 = 24;	// Added routing method parameter
static const int VRT_VERSION_25 = 25;	// Moves some variables from CompElement to Element
static const int VRT_VERSION_26 = 26;	// Reworked codes in Pin to allow 2 wires per grid point
static const int VRT_VERSION_27 = 27;	// Modified Element to store additional wire info
static const int VRT_VERSION_28 = 28;	// Added options for wires to share holes or cross/overlay
static const int VRT_VERSION_29 = 29;	// Added color to component definition
static const int VRT_VERSION_30 = 30;	// Added pin label alignment
static const int VRT_VERSION_31 = 31;	// New treatment of labels under part rotation and centering.
static const int VRT_VERSION_32 = 32;	// Added size of solder mask and silk screen pen
static const int VRT_VERSION_33 = 33;	// Added PCB track mode, board edge margin, PTH/NPTH option
static const int VRT_VERSION_34 = 34;	// Added 3rd dimension to grid, and display layer
static const int VRT_VERSION_35 = 35;	// Added via dimensions
static const int VRT_VERSION_36 = 36;	// Added thin tracks style
static const int VRT_VERSION_37 = 37;	// Added flag to enable/disable vias
static const int VRT_VERSION_38 = 38;	// Added Bourns 3362 series trimpots
static const int VRT_VERSION_39 = 39;	// Added custom pin/hole size on components
static const int VRT_VERSION_40 = 40;	// Allow 50 mil holes, and fix bug with non-wire hole use
static const int VRT_VERSION_41 = 41;	// Allow user-specified colors for nodes
static const int VRT_VERSION_42 = 42;	// Auto-fix corruption of old VRTs on load
static const int VRT_VERSION_43 = 43;	// Added vero numbers and vero letters
static const int VRT_VERSION_44 = 44;	// Added option to show close tracks
static const int VRT_VERSION_45 = 45;	// Added layer preference for each component pin
static const int VRT_VERSION_46 = 46;	// Added pad offsets for PCB mode
static const int VRT_VERSION_47 = 47;	// Added component attribute to allow flying wires
static const int VRT_VERSION_48 = 48;	// Added board setting to show/hide flying wires
static const int VRT_VERSION_49 = 49;	// Added vertical alignment flag to text boxes
static const int VRT_VERSION_50 = 50;	// Added background color property
static const int VRT_VERSION_51 = 51;	// Added pinlabels to component editor
static const int VRT_VERSION_52 = 52;	// Added thermal relief tag size
static const int VRT_VERSION_53 = 53;	// Added X-thermals flag
static const int VRT_VERSION_CURRENT = VRT_VERSION_53;
